// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	System interface for sound.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_unix.c,v 1.5 1997/02/03 22:45:10 b1 Exp $";

//#define DISABLE_MIDI

#include "z_zone.h"

#include "i_system.h"
#include "i_sound.h"
#include "m_argv.h"
#include "m_misc.h"
#include "w_wad.h"
#include "doomdef.h"
#include <math.h>

#include "mmus2mid.h"
#include "timidity.h"
#include "s_sound.h"

#include "pnobridge.h"



#define SHRT_MAX 32767
#define SHRT_MIN -32768

// Needed for calling the actual sound output.
static int SAMPLECOUNT = 1024;
static int SAMPLESIZE = 2; //stereo-2, mono-1


// MWM 2000-01-08: Sample rate in samples/second
int snd_samplerate=22050;

int pitched_sounds=0;

// The actual output device.
int audio_fd;

typedef struct {
  int lumpnum;
  // The channel step amount...
  unsigned int step;
  // ... and a 0.16 bit remainder of last step.
  unsigned int stepremainder;
  unsigned int samplerate;
  // The channel data pointers, start and end.
  unsigned char* data;
  unsigned char* enddata;
  // Time/gametic that the channel started playing,
  //  used to determine oldest, which automatically
  //  has lowest priority.
  // In case number of active sounds exceeds
  //  available channels.
  int starttime;
  // Hardware left and right channel volume lookup.
  int *leftvol_lookup;
  int *rightvol_lookup;
} channel_info_t;

channel_info_t channelinfo[MAX_CHANNELS];


// Pitch to stepping lookup, unused.
int steptable[256];

// Volume lookups.
int vol_lookup[128*256];

#include "SoundMgr.h"
static SndStreamRef snd_stream=NULL;
typedef struct {uint32 __r9;uint32 __r10;} reg_sys_t;
static reg_sys_t reg_sys;
int in_mix=0;
int loop_mode;

static MidiSong *song;

static int paused_song;

int g_music_mode;


/* cph 
 * stopchan
 * Stops a sound, unlocks the data 
 */

static void stopchan(int i)
{
  if (channelinfo[i].data) /* cph - prevent excess unlocks */
  {
    channelinfo[i].data=NULL;
//    W_UnlockLumpNum(channelinfo[i].lumpnum);
  }
}

//
// This function adds a sound to the
//  list of currently active sounds,
//  which is maintained as a given number
//  (eight, usually) of internal channels.
// Returns a handle.
//
int addsfx(sfxinfo_t *sfx, int channel)
{
  stopchan(channel);

  // We will handle the new SFX.
  // Set pointer to raw data.
  {
    int lumpnum = sfx->lumpnum;
    size_t len = W_LumpLength(lumpnum);

    /* Find padded length */
    len -= 8;
    channelinfo[channel].data = W_CacheLumpNum(lumpnum,PU_STATIC);
      
    /* Set pointer to end of raw data. */
    channelinfo[channel].enddata = channelinfo[channel].data + len - 1;
    channelinfo[channel].samplerate = (channelinfo[channel].data[3]<<8)+channelinfo[channel].data[2];
    channelinfo[channel].data += 8; /* Skip header */
    channelinfo[channel].lumpnum = lumpnum;
  }

  channelinfo[channel].stepremainder = 0;
  // Should be gametic, I presume.
  channelinfo[channel].starttime = gametic;

  return channel;
}

static void updateSoundParams(int handle, int volume, int seperation, int pitch)
{
  int slot = handle;
  int rightvol;
  int leftvol;
  int step = steptable[pitch];

#ifdef RANGECHECK
  if (handle>=MAX_CHANNELS)
  {
    I_Error("I_UpdateSoundParams: handle out of range");
    return;
  }
#endif
  // Set stepping
  // MWM 2000-12-24: Calculates proportion of channel samplerate
  // to global samplerate for mixing purposes.
  // Patched to shift left *then* divide, to minimize roundoff errors
  // as well as to use SAMPLERATE as defined above, not to assume 11025 Hz
  if (pitched_sounds)
    channelinfo[slot].step = step + (((channelinfo[slot].samplerate<<16)/snd_samplerate)-65536);
  else
    channelinfo[slot].step = ((channelinfo[slot].samplerate<<16)/snd_samplerate);
  
  // Separation, that is, orientation/stereo.
  //  range is: 1 - 256
  seperation += 1;

  // Per left/right channel.
  //  x^2 seperation,
  //  adjust volume properly.
  leftvol = volume - ((volume*seperation*seperation) >> 16);
  seperation = seperation - 257;
  rightvol= volume - ((volume*seperation*seperation) >> 16);  

  // Sanity check, clamp volume.
  if (rightvol < 0 || rightvol > 127)
    I_Error("rightvol out of bounds");
    
  if (leftvol < 0 || leftvol > 127)
    I_Error("leftvol out of bounds");
    
  // Get the proper lookup table piece
  //  for this volume level???
  channelinfo[slot].leftvol_lookup = &vol_lookup[leftvol*256];
  channelinfo[slot].rightvol_lookup = &vol_lookup[rightvol*256];
}

void I_UpdateSoundParams(int handle, int volume, int seperation, int pitch)
{
//  SDL_LockAudio();
  updateSoundParams(handle, volume, seperation, pitch);
//  SDL_UnlockAudio();
}

//
// SFX API
// Note: this was called by S_Init.
// However, whatever they did in the
// old DPMS based DOS version, this
// were simply dummies in the Linux
// version.
// See soundserver initdata().
//
void I_SetChannels()
{
  // Init internal lookups (raw data, mixing buffer, channels).
  // This function sets up internal lookups used during
  //  the mixing process. 
  int   i;
  int   j;
    
  int*  steptablemid = steptable + 128;
  
  // Okay, reset internal mixing channels to zero.
  for (i=0; i<MAX_CHANNELS; i++)
  {
    memset(&channelinfo[i],0,sizeof(channel_info_t));
  }

  // This table provides step widths for pitch parameters.
  // I fail to see that this is currently used.
  for (i=-128 ; i<128 ; i++)
    steptablemid[i] = (int)(pow(1.2, ((double)i/(64.0*snd_samplerate/11025)))*65536.0);
  
  
  // Generates volume lookup tables
  //  which also turn the unsigned samples
  //  into signed samples.
  for (i=0 ; i<128 ; i++)
    for (j=0 ; j<256 ; j++)
    {
      // proff - made this a little bit softer, because with
      // full volume the sound clipped badly
      vol_lookup[i*256+j] = (i*(j-128)*256)/191;
      //vol_lookup[i*256+j] = (i*(j-128)*256)/127;
    }
} 

 
void I_SetSfxVolume(int volume)
{
  // Identical to DOS.
  // Basically, this should propagate
  //  the menu/config file setting
  //  to the state variable used in
  //  the mixing.
  snd_SfxVolume = volume;
}

//
// Retrieve the raw data lump index
//  for a given SFX name.
//
int I_GetSfxLumpNum(sfxinfo_t* sfx)
{
  char namebuf[9];
  sprintf(namebuf, "ds%s", sfx->name);
  return W_GetNumForName(namebuf);
}

//
// Starting a sound means adding it
//  to the current list of active sounds
//  in the internal channels.
// As the SFX info struct contains
//  e.g. a pointer to the raw data,
//  it is ignored.
// As our sound handling does not handle
//  priority, it is ignored.
// Pitching (that is, increased speed of playback)
//  is set, but currently not used by mixing.
//
int I_StartSound(sfxinfo_t *sfx, int channel, int vol, int sep, int pitch, int priority)
{
  int handle;

  // UNUSED
  priority = 0;
  
  // Debug.
  //fprintf( stderr, "starting sound %d", id );
    
//  SDL_LockAudio();
  // Returns a handle (not used).
  handle = addsfx(sfx,channel);
#ifdef RANGECHECK
  if (handle>=MAX_CHANNELS)
  {
    I_Error("I_StartSound: handle out of range");
    return 0;
  }
#endif
  updateSoundParams(handle, vol, sep, pitch);
//  SDL_UnlockAudio();

  // fprintf( stderr, "/handle is %d\n", id );
    
  return handle;
}



void I_StopSound (int handle)
{
#ifdef RANGECHECK
  if (handle>=MAX_CHANNELS)
  {
    I_Error("I_StopSound: handle out of range");
    return;
  }
#endif
//  SDL_LockAudio();
  stopchan(handle);
//  SDL_UnlockAudio();
}


boolean I_SoundIsPlaying(int handle)
{
#ifdef RANGECHECK
  if (handle>=MAX_CHANNELS)
  {
    I_Error("I_SoundIsPlaying: handle out of range");
    return false;
  }
#endif
  return channelinfo[handle].data != NULL;
}


//
// This function loops all active (internal) sound
//  channels, retrieves a given number of samples
//  from the raw sound data, modifies it according
//  to the current (internal) channel parameters,
//  mixes the per channel samples into the given
//  mixing buffer, and clamping it to the allowed
//  range.
//
// This function currently supports only 16bit.
//

void I_UpdateSound(void *unused, UInt8 *stream, int len)
{
  // Mix current sound data.
  // Data, from raw sound, for right and left.
  register unsigned char sample;
  register int    dl;
  register int    dr;
  
  // Pointers in audio stream, left, right, end.
  signed short*   leftout;
  signed short*   rightout;
  signed short*   leftend;
  // Step in stream, left and right, thus two.
  int       step;

  // Mixing channel index.
  int       chan;
  int i;

  if (song&&loop_mode&&!paused_song)
  	if (!Timidity_Active()) Timidity_ReStart(song);
  
  if (song&&!paused_song&&Timidity_Active()) Timidity_PlaySome(stream,len);
  else for (i=0;i<len*2;i++) ((UInt16*)stream)[i]=0;

  // Left and right channel
  //  are in audio stream, alternating.
  leftout = (signed short *)stream;
  rightout = ((signed short *)stream)+1;
  step = 2;

  // Determine end, for left channel only
  //  (right channel is implicit).
  leftend = leftout + (len)*step;

  // Mix sounds into the mixing buffer.
  // Loop over step*SAMPLECOUNT,
  //  that is 512 values for two channels.
  while (leftout != leftend)
  {
    // Reset left/right value. 
    //dl = 0;
    //dr = 0;
    dl = *leftout;
    dr = *rightout;

    // Love thy L2 chache - made this a loop.
    // Now more channels could be set at compile time
    //  as well. Thus loop those  channels.
    for ( chan = 0; chan < numChannels; chan++ )
    {
      // Check channel, if active.
      if (channelinfo[chan].data)
      {
        // Get the raw data from the channel.
        // no filtering
        // sample = *channelinfo[chan].data;
        // linear filtering
        sample = (((unsigned int)channelinfo[chan].data[0] * (0x10000 - channelinfo[chan].stepremainder))
                + ((unsigned int)channelinfo[chan].data[1] * (channelinfo[chan].stepremainder))) >> 16;

        // Add left and right part
        //  for this channel (sound)
        //  to the current data.
        // Adjust volume accordingly.
        dl += channelinfo[chan].leftvol_lookup[sample];
        dr += channelinfo[chan].rightvol_lookup[sample];
        // Increment index ???
        channelinfo[chan].stepremainder += channelinfo[chan].step;
        // MSB is next sample???
        channelinfo[chan].data += channelinfo[chan].stepremainder >> 16;
        // Limit to LSB???
        channelinfo[chan].stepremainder &= 0xffff;

        // Check whether we are done.
        if (channelinfo[chan].data >= channelinfo[chan].enddata)
          stopchan(chan);
      }
    }
  
    // Clamp to range. Left hardware channel.
    // Has been char instead of short.
    // if (dl > 127) *leftout = 127;
    // else if (dl < -128) *leftout = -128;
    // else *leftout = dl;

    if (dl > SHRT_MAX)
      *leftout = SHRT_MAX;
    else if (dl < SHRT_MIN)
      *leftout = SHRT_MIN;
    else
      *leftout = (signed short)dl;

    // Same for right hardware channel.
    if (dr > SHRT_MAX)
      *rightout = SHRT_MAX;
    else if (dr < SHRT_MIN)
      *rightout = SHRT_MIN;
    else
      *rightout = (signed short)dr;

    // Increment current pointers in stream
    leftout += step;
    rightout += step;
  }
}

void I_ShutdownSound(void)
{   

    stub_SndStreamStop(snd_stream);
	stub_SndStreamDelete(snd_stream);   	   	   	

}

static Err tim_sndCallBack (void *userData,SndStreamRef stream, void *bufferZ,UInt32 frameCountZ)
{
	__asm 
	{
		stmfd  r13!,{r9,r10}
		ldr	   r9,[r0]
		ldr	   r10,[r0,#4]
	}	
	if (!in_mix){
		in_mix=1;		
		I_UpdateSound(NULL,bufferZ,frameCountZ);		
		in_mix=0;
	}
	__asm 
	{
		ldmfd  r13!,{r9,r10}
	}
	return errNone;
}

void I_InitSound()
{ 
	  	  		    
  // Secure and configure sound device first.

  /* Initialize variables */
  
 
  I_InitMusic();
    
  
  /* PALM Stuff */  
	asm {
  		ldr  r0, = reg_sys
	  	add	 r0,r0,r10
  		str	 r9,[r0]
	  	str	 r10,[r0,#4]
    	}     
	stub_SndStreamCreate(&snd_stream,sndOutput,snd_samplerate,(SAMPLESIZE==2?sndInt16Little:sndInt8),sndStereo,tim_sndCallBack,&reg_sys,SAMPLECOUNT);
	stub_SndStreamStart(snd_stream);
  
  // Finished initialization.
//  Sys_PrintF("I_InitSound: sound module ready\n");
}




//
// MUSIC API.
//


void I_PlaySong(int handle, int looping){
  if (song) {
  	  Sys_PrintF2("Starting music...please wait");
/*  	  {
  		  UInt32 bytes,max;
	  	  MemHeapFreeBytes ( 0, &bytes, &max );
		  Sys_PrintF("free ram before: %d/%d",bytes>>10,max>>10);
		  menu_info(" ");
	  }*/
      Timidity_StartFTR(song,CREATORID,FTR_MIDI_START_ID);            
      Timidity_SetVolume(snd_MusicVolume*4);
/*      {
  		  UInt32 bytes,max;
	  	  MemHeapFreeBytes ( 0, &bytes, &max );
		  Sys_PrintF("free ram after: %d/%d",bytes>>10,max>>10);
		  menu_info(" ");
	  }   */   
      


  }
  paused_song=0;
  loop_mode=looping;
}

void I_PauseSong (int handle){

  paused_song=1;
}

void I_ResumeSong (int handle)
{
  paused_song=0;
}

void I_StopSong(int handle)
{
    
	if (song) {
		Timidity_Stop();
		paused_song=2;
	}
}

void I_UnRegisterSong(int handle)
{
  if (!song){  
//  	I_Error("I_UnregisterSong : no song!");
  	return;
  }    
//Sys_PrintF("I_UnregisterSong");  
  Timidity_FreeSongFTR(song,CREATORID);
  song=NULL;
  remove("tmp.mid");
/*   {
   		UInt32 bytes,max;
	   	MemHeapFreeBytes ( 0, &bytes, &max );
	   	sprintf(gpstr,"ram : %d/%d",bytes/1024,max/1024);
	   	menu_message(gpstr);
	}*/
  
  
  
}

int I_RegisterSong(void *data,int len)
{
  MIDI *mididata;
  FILE *midfile;
  song=NULL;
  
  if (!g_music_mode) return 0;
  
  Sys_PrintF2("Loading music...please wait");

  midfile = fopen("tmp.mid", "wb");
  if ( midfile == NULL ) {
    I_Error("Couldn't write MIDI to %s\n", "tmp.mid");
    return 0;
  }
  /* Convert MUS chunk to MIDI? */
  if ( memcmp(data, "MUS", 3) == 0 )
  {
    UBYTE *mid;
    int midlen;

    mididata = ydk_malloc(sizeof(MIDI));
    mmus2mid(data, mididata, 89, 0);
    MIDIToMidi(mididata,&mid,&midlen);
    fwrite(mid, 1, midlen, midfile);
    free_mididata(mididata);
    ydk_free(mid);
    ydk_free(mididata);
  } else {
  	Sys_PrintF("already in mid format, size %d",MemPtrSize(data));
    fwrite(data, 1, len, midfile);
  }
  fclose(midfile);

  song=Timidity_LoadSong("/PALM/Programs/ZDoomZ/tmp.mid");  
  if (!song){
		I_Error("error at song loading");   		
		return 0;
  }
  return 1;
}

void I_SetMusicVolume(int volume)
{
  snd_MusicVolume = volume;
  if (Timidity_Active()) Timidity_SetVolume(snd_MusicVolume*4);
}

//
// MUSIC API.
//
void I_InitMusic(void)		{ 
	/*Timidity*/
//	Sys_PrintF("I_InitMusic");

#ifdef DISABLE_MIDI  
	g_music_mode=0;
#endif

	if (!g_music_mode) return;
    if (Timidity_Init(snd_samplerate,(SAMPLESIZE==2?AUDIO_S16LSB:AUDIO_S8),2,SAMPLECOUNT,"/PALM/Programs/Timidity/")){
    	menu_info("Error at tim init");
    	return;
    }
}
void I_ShutdownMusic(void)	{ 
	/*Timidity*/
//	Sys_PrintF("I_ShutdownMusic");
	if (!g_music_mode) return;	
	Timidity_Close();
}

