/*
	File:			ydkmain68k.c	

	Description:	Contains the 68K portion of the Palm ARMlet.  It sets up a form so that
					the UI works with the Palm UI, and it calls the ARMlet.

	Author:		

	Copyright:		Copyright  2003 by Tapwave, Inc.

	Disclaimer:		IMPORTANT:  This Tapwave software is provided by Tapwave, Inc. ("Tapwave").  Your 
					use is subject to and governed by terms and conditions of the Software Development 
					Kit Agreement ("SDK Agreement") between you and Tapwave.  If you have not entered 
					into Tapwaves standard SDK Agreement with Tapwave, you have no right or license 
					to use, reproduce, modify, distribute or otherwise exploit this Tapwave software.  
					You may obtain a copy of Tapwaves standard SDK Agreement by calling 650-960-1817
					or visiting Tapwave at http://www.tapwave.com/developers/. 

	Change History (most recent first):
				
*/

//#define TESTING_ONLY

#include <PalmOS.h>
#include <VFSMgr.h>
#include <StringMgr.h>
#include <PCENativeCall.h>
#include <SonyCLIE.h>
#include "ydk_pnoload.h"
#include "pnobridge.h"

#include <coretraps.h>

#include "TapWave/TwChars.h"
#include "stdfile.h"
#include "host68k.h"
#include "display68k.h"
#include "AppRsc.h"

#include <TwKeys.h>

#include <68k/System/HsExt.h>
#include <Common/System/HsKeyTypes.h>

// stupid TapWave...
#undef keyBitNavLeft
#undef keyBitNavRight
#undef keyBitNavSelect
#include "palm5way.h"

// isn't used anywhere else
#define VERSION_MAJOR "1"
#define VERSION_MINOR "7"
#define VERSION "1.7"
#define VERSION_DB 0x0107  //0xMMmm  M Major, m minor
#define WELCOME_MSG "Welcome to ZDoomZ 1.7"


#include <BtCommVdrv.h>
// Fix a bug in the Bluetooth SDK
#ifdef BtLibSdpUuidInitialize
#undef BtLibSdpUuidInitialize
#endif
#define BtLibSdpUuidInitialize(uuid, define, defineSize)  { (uuid).size = (defineSize); \
 MemMove((uuid).UUID, (define), (uuid).size); }


UInt16 n2PlayerRole = 0;
UInt16 nSerPortId = 0;
Boolean bSerPortOpened = false;

#include <NetMgr.h>
#define IGNORE_STDIO_STUBS
#include <unix/sys_socket.h>

#define INVALIDSOCKET ((NetSocketRef) -1)
#define IPPORT_USERRESERVED  5000

Boolean bNetLibOpened = false;
UInt16 netLibRefNum = 0;
NetSocketRef insocket = INVALIDSOCKET;
NetSocketRef outsocket = INVALIDSOCKET;
int	gblDoomPort = IPPORT_USERRESERVED + 0x1d;


// Define the minimum OS version we support (2.0 for now).
#define kPalmOS10Version	sysMakeROMVersion(1,0,0,sysROMStageRelease,0)

// collapsable input area
Int32 g_CollapseAPIVersion=0;
Err CollapseDisplayResizedEventCallback(SysNotifyParamType *notifyParamsP);
UInt16 SonySilkLibRefNum=0; // Sony's silk manager library

#define sonyVersion1 -1
#define sonyVersion2 -2

extern int scale;

int g_iwad_entries, g_pwad_entries;
char **g_iwad_listchoice, **g_pwad_listchoice;
int p_video_x, p_video_y, p_video_rot;
int gblPlayerMode = 0;
//int gblDeathMode = 0;
//int gblStartSkill = 0;
int gblDebugLog = 0;
//int gblNoMonster = 0;
//int gblRespawnMonster = 0;
int gblEpisode = 0;
int gblMap = 0;

char device_type;

char g_launch_pno;
extern UInt32 hw_flags;

typedef struct {UInt32 val;char *label;} t_hardkey;
t_hardkey hardkeys[]={
{keyBitPower,"Power"},
{keyBitPageUp,"Pg up"},
{keyBitPageDown,"Pg down"},
{keyBitHard1,"Hard1"},
{keyBitHard2,"Hard2"},
{keyBitHard3,"Hard3"},
{keyBitHard4,"Hard4"},
{keyBitCradle,"Cradle"},
{keyBitAntenna,"Antenna"},
{keyBitContrast,"Contrast"},
{keyBitNavUp,"Nav Up"},
{keyBitNavDown,"Nav Down"},
{keyBitNavLeft,"Nav Left"},
{keyBitNavRight,"Nav Right"},
{keyBitNavSelect,"Nav Select"},
{keyBitActionA,"Action A"},
{keyBitActionB,"Action B"},
{keyBitActionC,"Action C"},
{keyBitActionD,"Action D"},
{keyBitTriggerA,"Trig Left"},
{keyBitTriggerB,"Trig Right"},
{keyBitFunction,"Function"},
{keyBitLaunch,"Launch"},
{0,""}
};

// from "palm5way.h"
#define vchrNavChange           (vchrPalmMin + 3)

#define navBitsAll              0x001F
#define navChangeSelect         0x1000
#define navChangeBitsAll        0x1F00

#define IsFiveWayNavEvent(eventP)																		\
	(																												\
		 ((eventP)->data.keyDown.chr == vchrPageUp ||												\
		  (eventP)->data.keyDown.chr == vchrPageDown ||												\
		  (eventP)->data.keyDown.chr == vchrNavChange)												\
	&&																												\
		 (((eventP)->data.keyDown.keyCode & (navBitsAll | navChangeBitsAll)) != 0)			\
	)

#define NavSelectPressed(eventP)																			\
	(																												\
		IsFiveWayNavEvent(eventP)																			\
	&&																												\
		(((eventP)->data.keyDown.modifiers & autoRepeatKeyMask) == 0)							\
	&&																												\
		(((eventP)->data.keyDown.keyCode & (navBitsAll | navChangeBitsAll)) ==				\
																		 navChangeSelect)						\
	)


zdoomz_pref_t g_zdoomz_pref;

/**********************************/

Err CollapseDisplayResizedEventCallback( SysNotifyParamType *notifyParamsP );
	
void garbage_collect ( void );
void LoadZippedModuleFromCard(/*UInt16 id*/char *name);
UInt32 UnzipToFtr(char *name,UInt16 Ftr);
UInt32 LaunchPNO();

#define SET_FRM_COLORS() { \
			RGBColorType col;\
			Int16 cr,cg,cb,cr2,cg2,cb2;\
			cr=240;cg=240;cb=240;\
			col.r=cr;col.g=cg;col.b=cb;\
			UIColorSetTableEntry(UIFormFill,&col);\
			col.r>>=1;col.g>>=1;col.b>>=1;\
			UIColorSetTableEntry(UIFormFrame,&col);\
			cr2=220;cg2=200;cb2=250;\
			col.r=cr2;col.g=cg2;col.b=cb2;\
			UIColorSetTableEntry(UIObjectFill,&col);\
			col.r=(cr*0+cr2)>>1;\
			col.g=(cg*0+cr2)>>1;\
			col.b=(cb*0+cr2)>>1;\
			UIColorSetTableEntry(UIObjectFrame,&col);\
			cr=10;cg=10;cb=10;\
			col.r=cr;col.g=cg;col.b=cb;\
			UIColorSetTableEntry(UIObjectForeground,&col); \
			}


#define ASK_FOR_UPDATE \
	{EventType my_event; \
	my_event.eType=frmUpdateEvent;\
	EvtAddEventToQueue(&my_event);}


#define DRAW_LOGO(X,Y)\
	{\
		MemHandle hRsc=NULL;\
		hRsc = DmGetResource(bitmapRsc, LOGOBMP);\
		if (hRsc)\
		{	BitmapType* bitmapP = (BitmapType*) MemHandleLock(hRsc);\
			WinPaintBitmap(bitmapP, X, Y);\
			MemHandleUnlock(hRsc);\
			DmReleaseResource(hRsc);}\
	}


void drawList (Int16 i, RectangleType *bounds, Char **items, int num_entries, char **listchoice)
{
	Char *p,q[64];	
	Int16 j=0;
	Int16 wid;
	Boolean notruncated;
	IndexedColorType selectColor = UIColorGetTableEntryIndex(UIObjectSelectedForeground);
	IndexedColorType myColor=0;
	IndexedColorType itemColor = WinSetForeColor(myColor);	
	RGBColorType fg,bg;	
	
	WinSetForeColor(itemColor);	
	if (i>=num_entries) return;	
	p=listchoice[i];
		
	while (p[j++]);
	
	WinPushDrawState();	
    WinSetDrawMode(winPaint);
    
	j--;    
	
	switch (listchoice[i][0]) {
		case 3: //SLOT 2
			fg.r=100;fg.g=50;fg.b=200;
		break;
		case 2: //SLOT 1
			fg.r=120;fg.g=140;fg.b=40;
		break;
		case 1:
		default:  //SLOT 0 (internal)
			fg.r=50;fg.g=150;fg.b=70;
		break;
	}
	
	bg.r=255;bg.g=255;bg.b=255;       
	
    if (((p[j-1]=='P')||(p[j-1]=='p'))&&((p[j-2]=='I')||(p[j-2]=='i'))&&((p[j-3]=='Z')||(p[j-3]=='z'))) {
	    //FntSetFont(boldFont);
	}
	else{
	    UInt16 cr=(UInt16)(fg.r)*4/3;
	    UInt16 cg=(UInt16)(fg.g)*4/3;
	    UInt16 cb=(UInt16)(fg.b)*4/3;
	    if (cr>255) cr=255;if (cg>255) cg=255;if (cb>255) cb=255;
	    fg.r=cr;fg.b=cb;fg.b=cb;
//		fg.r<<=1;fg.g<<=1;fg.b<<=1;
	}   
	
	if (itemColor == selectColor) {
	    fg.r^=255;fg.g^=255;fg.b^=255;
		bg.r^=255;bg.g^=255;bg.b^=255;
	}
	        	
    WinSetTextColorRGB(&fg,NULL);    
//    WinSetBackColorRGB(&bg,NULL);


	wid=100;
	FntCharsInWidth(p,&wid,&j,&notruncated);
	if (!notruncated){
		if (j>63) {j=63;DEBUGS("drawList => truncated string with > 63chars");}
		p=q;//(Char*)MemPtrNew(j);
		MemMove(p,listchoice[i],j);
		p[j]=0;
		p[j-1]='.';p[j-2]='.';p[j-3]='.';
	}
	
	
	WinSetCoordinateSystem(kCoordinatesNative);
	
	WinPaintChars(p, j, bounds->topLeft.x * 2 / scale, bounds->topLeft.y * 2 / scale);
			
	WinSetDrawMode(winOverlay);
	
	if (itemColor == selectColor){
	fg.r<<=1; fg.g<<=1; fg.b<<=1;
    WinSetTextColorRGB(&fg,NULL);
	WinPaintChars(p, j, bounds->topLeft.x * 2 / scale - 1, bounds->topLeft.y * 2 / scale - 1);
	}		
//	if (!notruncated) MemPtrFree(p);
	WinPopDrawState();	
}

void	drawIWADList	(Int16	i,RectangleType	*bounds,Char		**items){
	drawList(i,bounds,items,g_iwad_entries,g_iwad_listchoice);
}
void	drawPWADList	(Int16	i,RectangleType	*bounds,Char		**items){
	drawList(i,bounds,items,g_pwad_entries,g_pwad_listchoice);
}


			
void StartGame()
{
	char *ptr;

	// Switch to fullscreen if choosed
	// and read current screen size
	// eventually turn on software landscape if necessary				
	if (g_zdoomz_pref.g_fullscreen)
	{
		hw_set_FullscreenGraf(1);
	}
	if (device_type & _ZODIAC_DEVICE_)
	{
		int res;

		if (res = tw_gfx_init())
		{
			DEBUGS("Cannot init Zodiac gfx");
			DEBUGU32(res);
			device_type &= ~_ZODIAC_DEVICE_;
		}
	}
	{
		UInt32 pitch, width, height;

    	hw_guess_screen_dimensions ( &width, &height, &pitch );			    	
    	if (g_zdoomz_pref.g_portrait)
    	{
    		p_video_x=320;
    		p_video_y=320;
    	}
    	else
    	{
	    	if (width==480)
	    	{
	    		p_video_x=480;
	    		p_video_y=320;
	    	}
	    	if (height==480)
	    	{
		    	p_video_x=480;
    			p_video_y=320;
    			p_video_rot=1;
	    	}
	    }
    }

	FtrSet(CREATORID, FTR_VIDEO_RESX_ID,p_video_x);
	FtrSet(CREATORID, FTR_VIDEO_RESY_ID,p_video_y);
	FtrSet(CREATORID, FTR_VIDEO_ROT_ID,p_video_rot);
	FtrSet(CREATORID, FTR_MUSIC_MODE_ID,g_zdoomz_pref.g_music_mode);
	FtrSet(CREATORID, FTR_RUMBLE_ID,g_zdoomz_pref.g_rumble);
	FtrSet(CREATORID, FTR_DEBUGFILE_ID, gblDebugLog);
	FtrSet(CREATORID, FTR_DEATHMODE_ID, g_zdoomz_pref.g_death_mode);
	FtrSet(CREATORID, FTR_STARTSKILL_ID, g_zdoomz_pref.g_start_skill);
	FtrSet(CREATORID, FTR_NOMONSTER_ID, g_zdoomz_pref.g_nomonster);
	FtrSet(CREATORID, FTR_RESPAWNMONSTER_ID, g_zdoomz_pref.g_respawn);
	FtrSet(CREATORID, FTR_EPISODE_ID, gblEpisode + 1);
	FtrSet(CREATORID, FTR_MAP_ID, gblMap + 1);

	FtrPtrNew(CREATORID, FTR_OPT_INPUTCONF_ID, INPUT_CONFIG_COUNT * sizeof(UInt32), (void**)&ptr);
	DmWrite(ptr,0,(void*)&(g_zdoomz_pref.g_inputconf[0]), INPUT_CONFIG_COUNT * sizeof(UInt32));

	FtrSet(CREATORID, FTR_SER_ROLE_ID, n2PlayerRole);
	FtrSet(CREATORID, FTR_SER_PORT_ID, nSerPortId);
	FtrSet(CREATORID, FTR_SER_OPEN_ID, bSerPortOpened);

	FtrSet(CREATORID, FTR_NET_OPEN_ID, bNetLibOpened);
	FtrSet(CREATORID, FTR_NET_IP_ID, g_zdoomz_pref.g_ipaddresses[0]);
	FtrSet(CREATORID, FTR_NET_LIB_ID, netLibRefNum);
	FtrSet(CREATORID, FTR_NET_INSOCK_ID, insocket);
	FtrSet(CREATORID, FTR_NET_OUTSOCK_ID, outsocket);
	FtrSet(CREATORID, FTR_NET_PORT_ID, gblDoomPort);
	
	LaunchPNO();

	if (bSerPortOpened)
		SrmClose(nSerPortId);
	if (bNetLibOpened)
	{
		Err err;
		NetLibSocketClose(netLibRefNum, outsocket, 20 * SysTicksPerSecond(), &err);
		NetLibSocketClose(netLibRefNum, insocket, 20 * SysTicksPerSecond(), &err);
		NetLibClose(netLibRefNum, false);
	}

	FtrUnregister(CREATORID, FTR_NET_OPEN_ID);
	FtrUnregister(CREATORID, FTR_NET_IP_ID);
	FtrUnregister(CREATORID, FTR_NET_LIB_ID);
	FtrUnregister(CREATORID, FTR_NET_INSOCK_ID);
	FtrUnregister(CREATORID, FTR_NET_OUTSOCK_ID);
	FtrUnregister(CREATORID, FTR_NET_PORT_ID);

	FtrUnregister(CREATORID, FTR_SER_ROLE_ID);
	FtrUnregister(CREATORID, FTR_SER_PORT_ID);
	FtrUnregister(CREATORID, FTR_SER_OPEN_ID);
				
	FtrPtrFree(CREATORID, FTR_OPT_INPUTCONF_ID);
	FtrUnregister(CREATORID, FTR_VIDEO_RESX_ID);
	FtrUnregister(CREATORID, FTR_VIDEO_RESY_ID);
	FtrUnregister(CREATORID, FTR_VIDEO_ROT_ID);
	FtrUnregister(CREATORID, FTR_MUSIC_MODE_ID);
	FtrUnregister(CREATORID, FTR_RUMBLE_ID);
	FtrUnregister(CREATORID, FTR_DEBUGFILE_ID);
	FtrUnregister(CREATORID, FTR_DEATHMODE_ID);
	FtrUnregister(CREATORID, FTR_STARTSKILL_ID);
	FtrUnregister(CREATORID, FTR_NOMONSTER_ID);
	FtrUnregister(CREATORID, FTR_RESPAWNMONSTER_ID);
	FtrUnregister(CREATORID, FTR_EPISODE_ID);
	FtrUnregister(CREATORID, FTR_MAP_ID);
	
	if (device_type & _ZODIAC_DEVICE_)
	{
		tw_gfx_close();
	}
	if (g_zdoomz_pref.g_fullscreen)
	{
		hw_set_FullscreenGraf(0);
	}
}


static void HandleLaunchButton(FormPtr frmP)
{
	Int16 idx;
	ListType *list;
	char *ftrptr;

	list = (ListType*) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, FORM_LISTFILE_ID));
	idx = LstGetSelection(list);
	if ((idx >= 0) && (idx < g_iwad_entries))
	{
		if (!g_zdoomz_pref.key_defined)
		{
			menu_inform("You forgot to setup controls.");
			menu_inform("HOME won't exit the game. ");
			menu_inform("So setup controls carefully.");
			ASK_FOR_UPDATE
		}
		else
		{
			StrCopy(g_zdoomz_pref.g_iwad_name,(g_iwad_listchoice[idx]));
			FtrPtrNew(CREATORID, FTR_WAD_FILENAME_ID, 256, (void **)&ftrptr);
			DmStrCopy(ftrptr, 0, g_zdoomz_pref.g_iwad_name);							
			g_launch_pno = 1;
		}
	}
	else
	{
		menu_inform("Please select a WAD file.");
		ASK_FOR_UPDATE
	}
	
	if (!g_launch_pno)
		return;

	list = (ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID) );
	idx = LstGetSelection(list);
	FtrPtrNew(CREATORID, FTR_PWAD_FILENAME_ID, 256, (void **)&ftrptr);
	if ((idx >= 0) && (idx < g_pwad_entries) && g_zdoomz_pref.g_pwad_active)
	{
		StrCopy(g_zdoomz_pref.g_pwad_name, g_pwad_listchoice[LstGetSelection(list)]);
	}
	else
	{
		g_zdoomz_pref.g_pwad_name[0] = 0;
	}
	DmStrCopy(ftrptr, 0, g_zdoomz_pref.g_pwad_name);

	if (gblPlayerMode == 3)
	{
		UInt16 inetErr = 0;
		Err err = errNone;
		UInt8 allInterfacesUp;
		UInt16 netIFErr;
		UInt32 **addrp;
		NetSocketAddrINType sockAddr;

		err = NetLibOpen(netLibRefNum, &inetErr);
		if (((err != errNone) && (err != netErrAlreadyOpen)) || (inetErr != errNone))
		{
			char tmp[64];

			if (err == netErrConfigNotFound)
				StrCopy(tmp, "Network not configured.");
			else if (err == netErrNoInterfaces)
				StrCopy(tmp, "No network interface defined.");
			else if (err == netErrPrefNotFound)
				StrCopy(tmp, "Network preference not found.");
			else if (err == netErrOutOfMemory)
				StrCopy(tmp, "Not enough memory.");
			else if (err == netErrConfigAliasErr)
				StrCopy(tmp, "Configuration error.");
			else if (inetErr == netErrNoInterfaces)
				StrCopy(tmp, "No interfaces defined.");
			else if (inetErr == netErrPrefNotFound)
				StrCopy(tmp, "Settings not found.");
			else if (inetErr == netErrUserCancel)
				StrCopy(tmp, "User cancelled.");
			else
				StrPrintF(tmp, "NetLibOpen: 0x%x inetErr: 0x%x", err, inetErr);
			DEBUGS(tmp);
		}
		else
		{
			// gblIPAddress
			NetLibConnectionRefresh(netLibRefNum, true, &allInterfacesUp, &netIFErr);
			outsocket = NetLibSocketOpen(netLibRefNum, netSocketAddrINET, netSocketTypeDatagram, netSocketProtoIPUDP, 20 * SysTicksPerSecond(), &err);
			if (outsocket == INVALIDSOCKET)
			{
				char tmp[16];
				if (err == netErrNoMoreSockets)
					StrCopy(tmp, "No more sockets.");
				else
					StrPrintF(tmp, "SocketOpen: 0x%x", err);
				DEBUGS(tmp);
				NetLibClose(netLibRefNum, false);
			}
			else
			{
				insocket = NetLibSocketOpen(netLibRefNum, netSocketAddrINET, netSocketTypeDatagram, netSocketProtoIPUDP, 20 * SysTicksPerSecond(), &err);
				if (insocket == INVALIDSOCKET)
				{
					char tmp[16];
					if (err == netErrNoMoreSockets)
						StrCopy(tmp, "No more sockets.");
					else
						StrPrintF(tmp, "SocketOpen: 0x%x", err);
					DEBUGS(tmp);
					NetLibSocketClose(netLibRefNum, outsocket, 20 * SysTicksPerSecond(), &err);
					NetLibClose(netLibRefNum, false);
				}
				else
				{
					MemSet(&sockAddr, sizeof(sockAddr), 0);
					sockAddr.family = AF_INET;
					sockAddr.port = htons(gblDoomPort);
//					sockAddr.addr = htonl(gblIPAddress);

					if (NetLibSocketBind(netLibRefNum, insocket, (NetSocketAddrType *) &sockAddr, sizeof(NetSocketAddrType), 20 * SysTicksPerSecond(), &err) == -1)
					{
						char tmp[64];
						if (err == 0x1214)
							StrCopy(tmp, "Bind closed by remote.");
						else if (err == 0x1212)
							StrCopy(tmp, "Bind timed out.");
						else
							StrPrintF(tmp, "Bind: 0x%x", err);
						DEBUGS(tmp);
						NetLibSocketClose(netLibRefNum, outsocket, 20 * SysTicksPerSecond(), &err);
						NetLibSocketClose(netLibRefNum, insocket, 20 * SysTicksPerSecond(), &err);
						NetLibClose(netLibRefNum, false);
					}
					else
					{
						bNetLibOpened = true;
					}
				}
			}
		}
	}
	else if (gblPlayerMode > 0)
	{
		SrmOpenConfigType config;
		BtVdOpenParams params;
		Err err;

		n2PlayerRole = gblPlayerMode - 1;

		if (n2PlayerRole == 0)
		{
			params.role = btVdClient;
			params.u.client.remoteDevAddr.address[0] = 0;
			params.u.client.remoteDevAddr.address[1] = 0;
			params.u.client.remoteDevAddr.address[2] = 0;
			params.u.client.remoteDevAddr.address[3] = 0;
			params.u.client.remoteDevAddr.address[4] = 0;
			params.u.client.remoteDevAddr.address[5] = 0;
			params.u.client.method = btVdUseUuidList;
			params.u.client.u.uuidList.len = 0;
		}
		else
		{
			params.role = btVdServer;
			BtLibSdpUuidInitialize(params.u.server.uuid, btLibSdpUUID_SC_SERIAL_PORT, btLibUuidSize16);
			params.u.server.name = "ZDoomZ";
		}
		params.authenticate = false;
		params.encrypt = false;

		config.baud = 57600;
		config.function = serFncUndefined;
		config.drvrDataP = &params;
		config.drvrDataSize = sizeof(params);
		config.sysReserved1 = 0;
		config.sysReserved2 = 0;

		err = SrmExtOpen(sysFileCVirtRfComm, &config, sizeof(config), &nSerPortId);
		if ((err == errNone) || (err == serErrAlreadyOpen))
			bSerPortOpened = true;
		else
			DEBUGS("Couldn't open BT port.");
	}

	garbage_collect();
	FrmGotoForm(GameForm);
}


/***********************************************************************
 *
 * FUNCTION:    BlankFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "BlankForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
Boolean BlankFormHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err			err = errNone;

	switch (eventP->eType) {
		case frmOpenEvent:{	
			ListType *list;
			static Boolean bFirst = true;
			
			SET_FRM_COLORS()
			FrmDrawForm(frmP);								
			
			list=(ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTFILE_ID) );
			LstSetDrawFunction (list, (ListDrawDataFuncPtr) drawIWADList);
			InitFileList(list,&g_iwad_entries,&g_iwad_listchoice,g_zdoomz_pref.g_iwad_name);
			
			list=(ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID) );
			LstSetDrawFunction (list, (ListDrawDataFuncPtr) drawPWADList);						
			InitFileList(list,&g_pwad_entries,&g_pwad_listchoice,g_zdoomz_pref.g_pwad_name);
			
			if (g_zdoomz_pref.g_pwad_active){
				FrmShowObject(frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID));
			}else{
				FrmHideObject(frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID));
			}
			
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,PWADCHECK),g_zdoomz_pref.g_pwad_active);
			
			if (!(g_zdoomz_pref.first)){
	   			menu_inform(WELCOME_MSG);
	   			g_zdoomz_pref.first=1;
	   		}
	   		if (bFirst)
	   		{
/*			menu_inform("2 Player version over BT");
			menu_inform("please send bug reports to");
			menu_inform("metaview@web.de");
*/			bFirst = false;
			}
			
			
			ASK_FOR_UPDATE

			handled = true;
			break;
		}
		case frmCloseEvent:{
			ListType *list;									
			Int16 idx;			
			list=(ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTFILE_ID) );
												
			if (list) {
			
				idx=LstGetSelection(list);
				if ((idx>=0)&&(idx<g_iwad_entries)){
					StrCopy(g_zdoomz_pref.g_iwad_name,(g_iwad_listchoice[idx]));
				}
			
				LstEraseList(list);
				LstSetListChoices(list,(char**)NULL,0);										
				if (g_iwad_listchoice) {
					int i;
					for (i=0;i<g_iwad_entries;i++)
						MemPtrFree(g_iwad_listchoice[i]);
					MemPtrFree(g_iwad_listchoice);
					g_iwad_listchoice=NULL;
				}				
				garbage_collect();
			}													
			list=(ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID) );
			if (list) {
			
				idx=LstGetSelection(list);
				if ((idx>=0)&&(idx<g_pwad_entries)&&g_zdoomz_pref.g_pwad_active){
					StrCopy(g_zdoomz_pref.g_pwad_name,(g_pwad_listchoice[idx]));
				} else g_zdoomz_pref.g_pwad_name[0]=0;
				
			
				LstEraseList(list);
				LstSetListChoices(list,(char**)NULL,0);
				if (g_pwad_listchoice) {
					int i;
					for (i=0;i<g_pwad_entries;i++)
						MemPtrFree(g_pwad_listchoice[i]);
					MemPtrFree(g_pwad_listchoice);
					g_pwad_listchoice=NULL;
				}				
				garbage_collect();
			}													
			break;
		}

		case ctlSelectEvent:{
			switch ( eventP->data.ctlSelect.controlID ) {			
				case PWADCHECK:
					g_zdoomz_pref.g_pwad_active=FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,PWADCHECK));
					if (g_zdoomz_pref.g_pwad_active){
						FrmShowObject(frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID));
					}else{
						FrmHideObject(frmP,FrmGetObjectIndex(frmP,FORM_LISTPATCHFILE_ID));
					}
					ASK_FOR_UPDATE
					handled=true;
					break;
				case LaunchBUTTON:
					HandleLaunchButton(frmP);
					handled = true;
					break;
				case ControlsBUTTON:
					FrmGotoForm(ConfigForm);
					handled=true;
					break;
				case OptionsBUTTON:
					FrmGotoForm(OptionsForm);
					handled=true;
					break;
				case InfoBUTTON:
					menu_inform("ZDoomz 1.7");
					menu_inform("Doom (c) ID Software");
					menu_inform("Palm Port : YoyoFR");
					menu_inform("http://yoyofr.fr.st");
					menu_inform("Enhanced: MetaView");
					menu_inform("http://www.metaviewsoft.de");
					menu_inform("Low-Res: Frankk");
					menu_inform("http://frankk.no-ip.com");
					ASK_FOR_UPDATE
					handled=true;
					break;
			}
			break;
		}
		case keyDownEvent:
			if ((eventP->data.keyDown.chr == vchrRockerCenter) || NavSelectPressed(eventP))
			{
				HandleLaunchButton(frmP);
				handled = true;
			}
			break;
		case winEnterEvent:	
		case winDisplayChangedEvent:	
		case frmUpdateEvent:
			{//ydk_WinScreenLock(winLockCopy);
				Char strDynMem[64];
				Char strDBCache[64];
				UInt32 max, free;

				FrmDrawForm(frmP);
				DRAW_LOGO(23,2)

				MemHeapFreeBytes(0, &free, &max);
				StrPrintF(strDynMem, "MEM %ld/%ld", max>>10, free>>10);
				WinDrawChars(strDynMem, StrLen(strDynMem), 10, 45);
				MemHeapFreeBytes(0x8001, &free, &max);
				StrPrintF(strDBCache, "DB %ld/%ld", max>>10, free>>10);
				WinDrawChars(strDBCache, StrLen(strDBCache), 90, 45);
			
			//ydk_WinScreenUnlock();
			}			
			handled=true;
			break;
	}
    
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    OptionsFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "OptionsForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
Boolean OptionsFormHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err			err = errNone;

	switch (eventP->eType) {
		case frmOpenEvent:{	
			ListType *list;
			Char tmp[8];
			
			SET_FRM_COLORS()
			FrmDrawForm(frmP);
			
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScreenList)), g_zdoomz_pref.g_portrait ? 1 : 0);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScreenTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScreenList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScreenList)))));
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,RUMBLECHECK), g_zdoomz_pref.g_rumble);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,MUSICCHECK), g_zdoomz_pref.g_music_mode);
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList)), gblPlayerMode);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList)))));
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DeathModeList)), g_zdoomz_pref.g_death_mode);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DeathModeList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DeathModeList)))));
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, StartSkillList)), g_zdoomz_pref.g_start_skill);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, StartSkillList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, StartSkillList)))));
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,DEBUGLOGCHECK),gblDebugLog);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,NOMONSTERCHECK), g_zdoomz_pref.g_nomonster);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,RESPAWNMONSTERCHECK), g_zdoomz_pref.g_respawn);
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, EpisodeList)), gblEpisode);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, EpisodeTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, EpisodeList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, EpisodeList)))));
			LstSetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MapList)), gblMap);
			CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MapTrigger)), LstGetSelectionText(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MapList)), LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MapList)))));
			
			if (LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList))) == 3)
			{
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
			}
			else if (LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList))) == 1)
			{
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
				FrmShowObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
			}
			else
			{
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
				FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
			}
			StrPrintF(tmp, "%d", g_zdoomz_pref.g_playerid);
			FldSetSelection(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerNumberField)), 0, FldGetTextLength(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerNumberField))));
			FldInsert(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerNumberField)), tmp, StrLen(tmp));
			StrPrintF(tmp, "%ld", (g_zdoomz_pref.g_ipaddresses[0] >> 24) & 0xff);
			FldSetSelection(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress1Field)), 0, FldGetTextLength(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress1Field))));
			FldInsert(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress1Field)), tmp, StrLen(tmp));
			StrPrintF(tmp, "%ld", (g_zdoomz_pref.g_ipaddresses[0] >> 16) & 0xff);
			FldSetSelection(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress2Field)), 0, FldGetTextLength(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress2Field))));
			FldInsert(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress2Field)), tmp, StrLen(tmp));
			StrPrintF(tmp, "%ld", (g_zdoomz_pref.g_ipaddresses[0] >> 8) & 0xff);
			FldSetSelection(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress3Field)), 0, FldGetTextLength(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress3Field))));
			FldInsert(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress3Field)), tmp, StrLen(tmp));
			StrPrintF(tmp, "%ld", g_zdoomz_pref.g_ipaddresses[0] & 0xff);
			FldSetSelection(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress4Field)), 0, FldGetTextLength(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress4Field))));
			FldInsert(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress4Field)), tmp, StrLen(tmp));

			ASK_FOR_UPDATE

			handled = true;
			break;
		}
		case frmCloseEvent:{
			Char *ptr;

			g_zdoomz_pref.g_ipaddresses[0] = 0;
			if (LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScreenList))) == 0)
			{
				g_zdoomz_pref.g_portrait = 0;
				g_zdoomz_pref.g_fullscreen = 1;
			}
			else
			{
				g_zdoomz_pref.g_portrait = 1;
				g_zdoomz_pref.g_fullscreen = 0;
			}
			gblPlayerMode = LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList)));
			g_zdoomz_pref.g_death_mode = LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DeathModeList)));
			g_zdoomz_pref.g_start_skill = LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, StartSkillList)));
			gblDebugLog = FrmGetControlValue (frmP,FrmGetObjectIndex(frmP, DEBUGLOGCHECK));
			g_zdoomz_pref.g_nomonster = FrmGetControlValue (frmP,FrmGetObjectIndex(frmP, NOMONSTERCHECK));
			g_zdoomz_pref.g_respawn = FrmGetControlValue (frmP,FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
			gblEpisode = LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, EpisodeList)));
			gblMap = LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MapList)));
			ptr = FldGetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerNumberField)));
			if (ptr)
				g_zdoomz_pref.g_playerid = StrAToI(ptr);
			n2PlayerRole = g_zdoomz_pref.g_playerid;
			ptr = FldGetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress1Field)));
			if (ptr)
				g_zdoomz_pref.g_ipaddresses[0] |= (StrAToI(ptr) << 24);
			ptr = FldGetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress2Field)));
			if (ptr)
				g_zdoomz_pref.g_ipaddresses[0] |= (StrAToI(ptr) << 16);
			ptr = FldGetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress3Field)));
			if (ptr)
				g_zdoomz_pref.g_ipaddresses[0] |= (StrAToI(ptr) << 8);
			ptr = FldGetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, IPAddress4Field)));
			if (ptr)
				g_zdoomz_pref.g_ipaddresses[0] |= StrAToI(ptr);
			break;
		}
		case ctlSelectEvent:{
			switch ( eventP->data.ctlSelect.controlID ) {			
				case InfoBUTTON:{
					FrmHelp(helpID_options);
					ASK_FOR_UPDATE
					handled = true;
					break;
				}
				case MUSICCHECK:{
					g_zdoomz_pref.g_music_mode = FrmGetControlValue (frmP, FrmGetObjectIndex(frmP, MUSICCHECK));
					handled = true;
					break;
				}
				case RUMBLECHECK:{
					g_zdoomz_pref.g_rumble = FrmGetControlValue (frmP, FrmGetObjectIndex(frmP, RUMBLECHECK));
					handled = true;
					break;
				}
				case OptionsFrmHOMEButton:
					FrmGotoForm(BlankForm);
					handled = true;
					break;
			}
			break;
		}
		case popSelectEvent:
		{
			if (eventP->data.popSelect.controlID == PlayerModeTrigger)
			{
				if (LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList))) == 3)
				{
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
				}
				else if (LstGetSelection (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, PlayerModeList))) == 1)
				{
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
					FrmShowObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
				}
				else
				{
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, DeathModeTrigger));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, StartSkillTrigger));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, NOMONSTERCHECK));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, RESPAWNMONSTERCHECK));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberLabel));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, PlayerNumberField));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress1Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress2Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress3Field));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Label));
					FrmHideObject(frmP, FrmGetObjectIndex(frmP, IPAddress4Field));
				}
			}
			break;
		}
		case keyDownEvent:
			if ((eventP->data.keyDown.chr == vchrRockerCenter) || NavSelectPressed(eventP))
			{
				FrmGotoForm(BlankForm);
				handled = true;
			}
			break;
		case winEnterEvent:	
		case winDisplayChangedEvent:	
		case frmUpdateEvent:
			FrmDrawForm(frmP);
			DRAW_LOGO(23,2)
			WinDrawLine(4, 83, 156, 83);
			handled = true;
			break;
	}
    
	return handled;
}


Boolean GameFormHandleEvent(EventType * eventP)
{
	Boolean         handled = false;
	FormType       *frmP = FrmGetActiveForm();
	UInt32          flags, pinMgrVersion;
	Err             err = errNone;

	switch (eventP->eType)
	{
		case frmOpenEvent:
			FrmDrawForm(frmP);
			handled = true;
			break;

		case winEnterEvent:
		case winDisplayChangedEvent:
			{
				if (g_CollapseAPIVersion != 0)
				{
					Coord         drawAreaHeightCoords, drawAreaWidthCoords;
					Int16         offsetCoordsX, offsetCoordsY;
					RectangleType formBounds;
					WinHandle     frmWinH;

					WinGetDisplayExtent( &drawAreaWidthCoords, &drawAreaHeightCoords );
HostTraceOutputTL(1, "Display: %d %d", drawAreaWidthCoords, drawAreaHeightCoords);

					frmWinH = FrmGetWindowHandle(frmP);
					WinGetBounds(frmWinH, &formBounds);
HostTraceOutputTL(1, "Form: %d %d %d %d", formBounds.topLeft.x, formBounds.topLeft.y, formBounds.extent.x, formBounds.extent.y);
				
					//for full screen windows topLeft.y is simply 0, or 1 for modal dialogs
					offsetCoordsX = drawAreaWidthCoords - formBounds.extent.x - formBounds.topLeft.x;
					offsetCoordsY = drawAreaHeightCoords - formBounds.extent.y - formBounds.topLeft.y;

					if (offsetCoordsX || offsetCoordsY)
					{
						formBounds.extent.x += offsetCoordsX;
						formBounds.extent.y += offsetCoordsY;

						// Change current draw window to match the display extent (app's drawing area)
						WinSetBounds( frmWinH, &formBounds );
					}
				}
			}
			handled = true;
			break;

		case nilEvent:
			StartGame();
			{
				// this is the end for you my friend...
				// until we can recover all memory allocated by the emulation core
				EventType newEvent;
				MemSet(&newEvent, sizeof(newEvent), 0);
				newEvent.eType = keyDownEvent;
				newEvent.data.keyDown.chr = launchChr;
				newEvent.data.keyDown.modifiers = commandKeyMask;
				EvtAddEventToQueue (&newEvent);
			}
			handled = true;
			break;
	}

	return handled;
}


void ConfigFrmUpdateControls(FormType *frmP){
	Char strtmp[64],strtmp2[64];
	FieldType *field;
	UInt16	i,j,k;
	Char g_inputfld[2000];
	g_inputfld[0]=0;
	for (i=0; i<INPUT_CONFIG_COUNT; i++)
	{
		if (is_treo() && (g_zdoomz_pref.g_inputconf[i] & 0x80000000))
			StrPrintF(strtmp, "key: '%c'", (char) (g_zdoomz_pref.g_inputconf[i] & 0xff));
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_TR)
			StrCopy(strtmp, "S.TopRight");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_TM)
			StrCopy(strtmp, "S.TopMid");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_TL)
			StrCopy(strtmp, "S.TopLeft");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_BR)
			StrCopy(strtmp, "S.BottomRight");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_BM)
			StrCopy(strtmp, "S.BottomMid");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_BL)
			StrCopy(strtmp, "S.BottomLeft");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_ML)
			StrCopy(strtmp ,"S.MidLeft");
		else if (g_zdoomz_pref.g_inputconf[i] == BTN_MR)
			StrCopy(strtmp, "S.MidRight");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00010000)
			StrCopy(strtmp, "RockerUp");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00020000)
			StrCopy(strtmp, "RockerDn");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00040000)
			StrCopy(strtmp, "RockerLt");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00080000)
			StrCopy(strtmp, "RockerRt");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00100000)
			StrCopy(strtmp, "RockerCt");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00800000)
			StrCopy(strtmp, "Launcher");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x04000000)
			StrCopy(strtmp, "TriggerA");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x08000000)
			StrCopy(strtmp, "TriggerB");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x10000000)
			StrCopy(strtmp, "Blue    ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x20000000)
			StrCopy(strtmp, "Yellow  ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x40000000)
			StrCopy(strtmp, "Green   ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x80000000)
			StrCopy(strtmp, "Red     ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00008000)
			StrCopy(strtmp, "Function");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x01000000)
			StrCopy(strtmp, "NavLeft ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x02000000)
			StrCopy(strtmp, "NavRight");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000002)
			StrCopy(strtmp, "PageUp  ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000004)
			StrCopy(strtmp, "PageDown");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000008)
			StrCopy(strtmp, "Cal.    ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000010)
			StrCopy(strtmp, "PhoneBk ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000020)
			StrCopy(strtmp, "ToDo    ");
		else if (g_zdoomz_pref.g_inputconf[i] & 0x00000040)
			StrCopy(strtmp, "Memo    ");
		else if (g_zdoomz_pref.g_inputconf[i] == 0x00000000)
			StrCopy(strtmp, "None    ");
		else
			StrPrintF(strtmp,"%lx",g_zdoomz_pref.g_inputconf[i]);
		StrCat(g_inputfld,strtmp);					
		StrCat(g_inputfld,"\n");
	}
	field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
	FldSetTextPtr(field,g_inputfld);
	FldRecalculateField (field,1);
}

/***********************************************************************
 *
 * FUNCTION:    ConfigFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "ConfigForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
Boolean ConfigFormHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err			err = errNone;

	switch (eventP->eType) {
		case frmOpenEvent:{	
			ListType *list;
			
			SET_FRM_COLORS()
			FrmDrawForm(frmP);		
			
			ConfigFrmUpdateControls(frmP);
			ASK_FOR_UPDATE

			handled = true;
			break;
		}
		case frmCloseEvent:
		{
			/*SAVE PREF*/
			PrefSetAppPreferences(CREATORID,0x0001,(short)VERSION_DB,(void*)NULL,(short)0,(unsigned char)0);
			PrefSetAppPreferences(CREATORID,0x0001,(short)VERSION_DB,(void*)&g_zdoomz_pref,(short)sizeof(g_zdoomz_pref),(unsigned char)0);
			break;
		}
		case ctlSelectEvent:
		{
			switch ( eventP->data.ctlSelect.controlID )
			{
				case ConfFrmHOMEButton:
					FrmGotoForm(BlankForm);					
					handled=true;
					break;			
				case ConfFrmASSIGNButton:
				{
					ListType *list;
					UInt32 i, j, k, old_mask;
					UInt32 newtreomask[3] = {0}, oldtreomask[3] = {0};
					FieldType *field;
					list = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConfFrmINPUTList));
					i = LstGetSelection(list);
					ydk_FadeScreen();
					menu_messageXY(0,0,80,60,"TL",0x0AC00A);
					menu_messageXY(240,0,80,60,"TR",0x0AC00A);
 					menu_messageXY(0,120,80,60,"ML",0x0AC00A);
					menu_messageXY(240,120,80,60,"MR",0x0AC00A);
					menu_messageXY(0,260,80,60,"BL",0x0AC00A);
					menu_messageXY(240,260,80,60,"BR",0x0AC00A);
					menu_messageXY(120,0,80,60,"TM",0x0AC00A);
					menu_messageXY(120,260,80,60,"BM",0x0AC00A);
					
					menu_messageXY(158-60-40,204,80,30,"Cancel",0x80600A);
					menu_messageXY(158+60-40,204,80,30,"None",0x80600A);
					menu_messageXY(16,84,288,32,"Press a button or hit screen",0xC89664);
					old_mask = KeySetMask(~keyBitsAll);
					if (is_treo())
					{
						HsKeySetMaskExt(newtreomask, oldtreomask);
/*
						HsKeyEnableKey(keyBitExt2Symbol, false);
						HsKeyEnableKey(keyBitExt2Launcher, false);
						HsKeyEnableKey(keyBitExt2CmdMenu, false);
						HsKeyEnableKey(keyBitExt2VolumeUp, false);
						HsKeyEnableKey(keyBitExt2VolumeDown, false);
*/
					}
					if (menu_waitBtn(1,0,&k))
					{
						int l;
						if (k)
						{
							g_zdoomz_pref.key_defined=1;
							for (l=0; l<INPUT_CONFIG_COUNT; l++)
								if (g_zdoomz_pref.g_inputconf[l]==k) g_zdoomz_pref.g_inputconf[l]=0;
						}
						g_zdoomz_pref.g_inputconf[i]=k;
						ConfigFrmUpdateControls(frmP);
					}
					if (is_treo())
					{
						HsKeySetMaskExt(oldtreomask, newtreomask);
/*
						HsKeyEnableKey(keyBitExt2Symbol, true);
						HsKeyEnableKey(keyBitExt2Launcher, true);
						HsKeyEnableKey(keyBitExt2CmdMenu, true);
						HsKeyEnableKey(keyBitExt2VolumeUp, true);
						HsKeyEnableKey(keyBitExt2VolumeDown, true);
*/
					}
					KeySetMask(old_mask/*keyBitsAll*/);
					
					if (i<LstGetNumberOfItems(list)-1) i++;
					LstSetSelection(list,i);
					
					field = (FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
					FldSetScrollPosition(field, 0);
					FldScrollField(field,LstGetTopItem(list), winDown);
					ASK_FOR_UPDATE
					break;
				}								
			}
		}
		case penUpEvent:
		{
			ListType *list;
			FieldType *field = (FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
			FldSetScrollPosition(field,0);
			list = (ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTList) );
			FldScrollField(field,LstGetTopItem(list),winDown);
			ASK_FOR_UPDATE
			break;
		}
		case keyDownEvent:
			if ((eventP->data.keyDown.chr == vchrRockerCenter) || NavSelectPressed(eventP))
			{
				FrmGotoForm(BlankForm);
				handled = true;
			}
			else if (eventP->data.keyDown.chr == vchrPageDown)
			{
				ListType *list = (ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTList) );
				FieldType *field = (FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
				if (LstGetTopItem(list) < (LstGetNumberOfItems(list) - LstGetVisibleItems(list)))
					LstSetTopItem(list, LstGetTopItem(list)+1);
				FldSetScrollPosition(field,0);
				FldScrollField(field,LstGetTopItem(list),winDown);
				ASK_FOR_UPDATE
			}
			else if (eventP->data.keyDown.chr == vchrPageUp)
			{
				ListType *list = (ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTList) );
				FieldType *field = (FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
				if (LstGetTopItem(list) > 0)
					LstSetTopItem(list, LstGetTopItem(list)-1);
				FldSetScrollPosition(field,0);
				FldScrollField(field,LstGetTopItem(list),winDown);
				ASK_FOR_UPDATE
			}
			break;
		case winEnterEvent:	
		case winDisplayChangedEvent:	
		case frmUpdateEvent:
			{//ydk_WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);
			DRAW_LOGO(23,2)			
			
			//ydk_WinScreenUnlock();
			}			
			handled=true;
			break;
			
		
	}
    
	return handled;
}




/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 * 				handler for the form loaded.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should not be passed
 *				to a higher level handler.
 *
 ***********************************************************************/
Boolean AppHandleEvent(EventType* eventP)
{
	UInt16 formId;
	FormType * frmP= FrmGetActiveForm();
	WinHandle formWinH;
	
	if (eventP->eType == frmLoadEvent)
	{
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		formWinH = FrmGetWindowHandle(frmP);
		if (formId == GameForm)
		{
			if (g_CollapseAPIVersion > 0)
			{
				WinSetConstraintsSize(formWinH, 80, 160, 240, 160, 160, 240);
				FrmSetDIAPolicyAttr(frmP, frmDIAPolicyCustom);
				PINSetInputTriggerState(pinInputTriggerDisabled);
				PINSetInputAreaState(pinInputAreaClosed);
				hw_set_FullscreenGraf(1);
			}
			else if (g_CollapseAPIVersion < 0)
			{
				SilkLibEnableResize(SonySilkLibRefNum);
				SilkLibResizeDispWin(SonySilkLibRefNum, silkResizeMax);
				if (g_CollapseAPIVersion == sonyVersion2)
				{
					VskSetState(SonySilkLibRefNum, vskStateEnable, (vskResizeVertically|vskResizeHorizontally));
					VskSetState(SonySilkLibRefNum, vskStateResize, vskResizeNone);
					VskSetState(SonySilkLibRefNum, vskStateEnable, 0);
				}
			}
		}
		else
		{
			if (g_CollapseAPIVersion > 0)
			{
				WinSetConstraintsSize(formWinH, 160, 160, 160, 160, 160, 160);
				FrmSetDIAPolicyAttr(frmP, frmDIAPolicyStayOpen);
				PINSetInputTriggerState(pinInputTriggerDisabled);
				PINSetInputAreaState(pinInputAreaOpen);
				hw_set_FullscreenGraf(0);
			}
			else if (g_CollapseAPIVersion < 0)
			{
				SilkLibEnableResize(SonySilkLibRefNum);
				SilkLibResizeDispWin(SonySilkLibRefNum, silkResizeNormal);
				SilkLibDisableResize(SonySilkLibRefNum);
				if (g_CollapseAPIVersion == sonyVersion2)
				{
					VskSetState(SonySilkLibRefNum, vskStateEnable, (vskResizeVertically|vskResizeHorizontally));
					VskSetState(SonySilkLibRefNum, vskStateResize, vskResizeMax);
					VskSetState(SonySilkLibRefNum, vskStateEnable, 0);
				}
			}
		}

		/* 
		 * Set the event handler for the form.  The handler of the
		 * currently active form is called by FrmHandleEvent each
		 * time is receives an event. 
		 */
		switch (formId)
		{
			case BlankForm:
				FrmSetEventHandler( frmP, BlankFormHandleEvent );
				break;
			case GameForm:
				FrmSetEventHandler( frmP, GameFormHandleEvent );
				break;
			case ConfigForm:
				FrmSetEventHandler( frmP, ConfigFormHandleEvent );
				break;
			case OptionsForm:
				FrmSetEventHandler( frmP, OptionsFormHandleEvent );
				break;
		}
		return true;
	}		
	return false;
}

void AppEventLoop(void)
{
	UInt16 error;
	EventType event;

	do
	{
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, 20);

		if (! SysHandleEvent(&event))
		{
			if (! AppHandleEvent(&event))
			{
				if (! MenuHandleEvent(0, &event, &error))
				{
					FrmDispatchEvent(&event);
				}
			}
		}
	} while (event.eType != appStopEvent);
}

void AppStop(void)
{
/*SAVE PREF*/
	PrefSetAppPreferences(CREATORID,0x0001,(short)VERSION_DB,(void*)NULL,(short)0,(unsigned char)1);
	PrefSetAppPreferences(CREATORID,0x0001,(short)VERSION_DB,(void*)&g_zdoomz_pref,(short)sizeof(g_zdoomz_pref),(unsigned char)1);
	FrmCloseAllForms();
}


void CheckDirs(void)
{
	UInt16 volRefNum;			
	UInt32 volIterator; 
	Err err;
	
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err)
		{
			VFSDirCreate(volRefNum,VFSPATH);
		}
	}
}


void InitPrefDefaultValues()
{
	MemSet(&g_zdoomz_pref, 0, sizeof(g_zdoomz_pref));

	if (device_type & _ZODIAC_DEVICE_)
	{
		g_zdoomz_pref.g_fullscreen = 1;
		g_zdoomz_pref.g_portrait = 0;
	}
	else
	{
		g_zdoomz_pref.g_portrait = 1;
		g_zdoomz_pref.g_fullscreen = 0;
	}

#define	keyTwBitFunction        0x00008000
	if (has_5way())
	{
		// FIXME: the E2 is supposed to have 5way, but it is actually rocker?
		g_zdoomz_pref.g_inputconf[0] = keyBitPageUp;
		g_zdoomz_pref.g_inputconf[1] = keyBitPageDown;
		g_zdoomz_pref.g_inputconf[2] = keyBitNavLeft;
		g_zdoomz_pref.g_inputconf[3] = keyBitNavRight;
		g_zdoomz_pref.g_inputconf[4] = BTN_BL;
		g_zdoomz_pref.g_inputconf[5] = 0;
		g_zdoomz_pref.g_inputconf[6] = keyBitHard2;
		g_zdoomz_pref.g_inputconf[7] = keyBitHard3;
		g_zdoomz_pref.g_inputconf[8] = keyBitNavSelect;
		g_zdoomz_pref.g_inputconf[9] = keyBitHard1;
		g_zdoomz_pref.g_inputconf[10] = keyBitHard4;
		g_zdoomz_pref.g_inputconf[11] = BTN_BM;
		g_zdoomz_pref.g_inputconf[12] = 1;
		g_zdoomz_pref.g_inputconf[13] = BTN_BR;
		g_zdoomz_pref.g_inputconf[14] = BTN_MR;
		g_zdoomz_pref.g_inputconf[15] = BTN_TR;
	}
	else if (is_zodiac())
	{
		g_zdoomz_pref.g_inputconf[0] = keyBitPageUp;
		g_zdoomz_pref.g_inputconf[1] = keyBitPageDown;
		g_zdoomz_pref.g_inputconf[2] = keyBitRockerLeft;
		g_zdoomz_pref.g_inputconf[3] = keyBitRockerRight;
		g_zdoomz_pref.g_inputconf[4] = keyBitHard1;
		g_zdoomz_pref.g_inputconf[5] = 0;
		g_zdoomz_pref.g_inputconf[6] = keyBitTriggerA;
		g_zdoomz_pref.g_inputconf[7] = keyBitTriggerB;
		g_zdoomz_pref.g_inputconf[8] = keyBitRockerCenter;
		g_zdoomz_pref.g_inputconf[9] = keyBitHard2;
		g_zdoomz_pref.g_inputconf[10] = keyTwBitFunction;
		g_zdoomz_pref.g_inputconf[11] = keyBitHard3;
		g_zdoomz_pref.g_inputconf[12] = keyBitLaunch;
		g_zdoomz_pref.g_inputconf[13] = keyBitHard4;
		g_zdoomz_pref.g_inputconf[14] = BTN_MR;
		g_zdoomz_pref.g_inputconf[15] = BTN_TR;
	}
	else
	{
		g_zdoomz_pref.g_inputconf[0] = keyBitPageUp;
		g_zdoomz_pref.g_inputconf[1] = keyBitPageDown;
		g_zdoomz_pref.g_inputconf[2] = keyBitRockerLeft;
		g_zdoomz_pref.g_inputconf[3] = keyBitRockerRight;
		g_zdoomz_pref.g_inputconf[4] = BTN_BL;
		g_zdoomz_pref.g_inputconf[5] = 0;
		g_zdoomz_pref.g_inputconf[6] = keyBitHard2;
		g_zdoomz_pref.g_inputconf[7] = keyBitHard3;
		g_zdoomz_pref.g_inputconf[8] = keyBitRockerCenter;
		g_zdoomz_pref.g_inputconf[9] = keyBitHard1;
		g_zdoomz_pref.g_inputconf[10] = keyBitHard4;
		g_zdoomz_pref.g_inputconf[11] = BTN_BM;
		g_zdoomz_pref.g_inputconf[12] = 1;
		g_zdoomz_pref.g_inputconf[13] = BTN_BR;
		g_zdoomz_pref.g_inputconf[14] = BTN_MR;
		g_zdoomz_pref.g_inputconf[15] = BTN_TR;
	}
}


UInt32 query_heap_free ( void )
{
	UInt16 heapid = 0;
	UInt32 max, bytes;

	MemHeapFreeBytes ( heapid, &bytes, &max );

	return ( max );
}


void garbage_collect ( void ) {

  /* on failure, compact memory */
  //MemHeapCompact ( MemPtrHeapID ( true_head() ) );

//#ifdef RSRC
  MemPtr ptr=MemPtrNew(8);	
  UInt16 heap_id=MemPtrHeapID(ptr);
  MemPtrFree(ptr);
//  MemHeapCompact ( heap_id );
//#endif

  return;
}


void LoadZippedModuleFromCard(/*UInt16 id*/char *name)
{	
  UInt32 processor;
  UInt32 flags = 0;
  ControlType *c;
  FormPtr frmP = FrmGetActiveForm();
  UInt8 filetype;
  Err res;
  UInt16 initDelay,old_initDelay,old_initDblDelay;
  UInt16 period,old_period;
  UInt32 old_mask;
  Boolean queueAhead,old_queueAhead;
  MemPtr pnoPtr,ptr;
  PnoDescriptor pno;
/*  char *name;  
  MemHandle memh;
  FILE *f;
  name="tmp.pno";
  memh = DmGetResource('YDKC', id);
  if (!memh) {DEBUGS("Error while accessing pno resource");return;}
  ptr = MemHandleLock(memh);
  f=fopen("tmp.pno","wb");
  if (!f) {DEBUGS("Cannot write temp pno"); return;}
  fwrite(ptr,1,MemPtrSize(ptr),f);
  fclose(f);
  MemHandleUnlock(memh);
	DmReleaseResource(memh);
  */


//  StrPrintF(name,"%1d:%s%s",palm_vfs_iter(),VFSPATH,_name);
#ifdef TESTING_ONLY	
	return;
#endif	  
  // does emu module exist?
  {
    FILE *f = fopen ( name, "r" );
    if ( ! f ) {
      DEBUGS ( "The emulation module to run this game is not present." );
      DEBUGS (name);
      return;
    }
    fclose ( f );
  }
  // clear background
  {
	UInt32 pitch, width, height;
	UInt16 *screen = (UInt16 *) hw_query_vram();

	hw_guess_screen_dimensions ( &width, &height, &pitch );			    	
	MemSet(screen, height * 2 * pitch, 0);
  }
  // call the PACE Native Object
  FtrGet(sysFtrCreator, sysFtrNumProcessorID, &processor);
  // rearrange memory
  garbage_collect();
  // fire up pno
  if ( sysFtrNumProcessorIsARM ( processor ) )
  {
    pnobridge_t *bridge = (pnobridge_t*) MemPtrNew ( sizeof(pnobridge_t) );
    MemHandle mh;
    UInt8 *tmpPtr;
    UInt32 pitch, width, height;
    Err error;    

    UnzipToFtr(name,FTR_PNO);
	FtrGet(CREATORID, FTR_PNO, (void*)&pnoPtr);		
	
	hw_guess_screen_dimensions ( &width, &height, &pitch );
    bridge -> vram = (UInt32) hw_query_vram();
    bridge -> eventf = (UInt32) host_event_handler;
    bridge -> display_width = width;
   	bridge -> display_height = height;
    bridge -> display_pitch = pitch;
    
    bridge -> param1 = NULL;
    
    bridge -> param2 = NULL;
    bridge -> flags = (hw_flags&hf_pin_present?BF_HASPIN:0)|
    				(device_type&_ZODIAC_DEVICE_?BF_ZODIAC:0)|
    				(device_type&_TREO_DEVICE_?BF_TREO:0)|
    				(device_type&_HIRES_PLUS_DEVICE_?BF_HIRESPLUS:0);
	
    res=ydk_PnoLoad ( &pno, pnoPtr );
    if (res)
    {
		DEBUGS("Error while loading PNO");
		DEBUGU32(res);
		MemPtrFree(bridge);
	    FtrPtrFree ( CREATORID, FTR_PNO ); 
		return;    	
    }
    else
    {
		old_mask=KeySetMask(~keyBitsAll);		
		initDelay=slowestKeyDelayRate;
		period=slowestKeyPeriodRate;
		queueAhead=false;
		KeyRates(false,&old_initDelay,&old_period,&old_initDblDelay,&old_queueAhead);
		KeyRates(true,&initDelay,&period,&initDelay,&queueAhead);
		//read memory to flush D cache
		FLUSH_D_CACHE	

	    res = ydk_PnoCall ( &pno, bridge );        
		{
			EventType myevent;
			for (;;)
			{
				EvtGetEvent(&myevent,0);
				if (myevent.eType==nilEvent) break;
				//SysHandleEvent(&myevent);
			}
		}
		KeySetMask(old_mask/*keyBitsAll*/);
		KeyRates(true,&old_initDelay,&old_period,&old_initDblDelay,&old_queueAhead);
	
		//unload pno
		res=ydk_PnoUnload ( &pno );
	    if (res) {DEBUGS("Error while loading PNO, probably not enough free ram.");DEBUGS(name);}
	    MemPtrFree(bridge);
    
	    FtrPtrFree ( CREATORID, FTR_PNO ); 
		// rearrange memory
		garbage_collect();
	}
  } // is arm and rom picked
}

UInt32 UnzipToFtr(char *name,UInt16 Ftr){
	MemHandle memh;
	MemPtr ptr;
	PnoDescriptor pno;
	UInt32 res;
	pnobridge_t *bridge;	
	UInt32 pitch, width, height;
	
	memh = DmGetResource('YDKC', PNO_ZLIB);
	if (!memh) {DEBUGS("Error while accessing zlib resource");return 1;}
	ptr = MemHandleLock(memh);
	if (!ptr){DEBUGS("Error while locking zlib resource");return 2;}	
	res=ydk_PnoLoad ( &pno, ptr );
    FtrPtrFree ( CREATORID, FTR_PNO ); 
    if (res) {DEBUGS("Error while loading zlib pno");return 3;}
	//read memory to flush D cache
	FLUSH_D_CACHE

    bridge = (pnobridge_t*) MemPtrNew ( sizeof(pnobridge_t) );
    hw_guess_screen_dimensions ( &width, &height, &pitch );
    bridge -> vram = (UInt32) hw_query_vram();
    bridge -> eventf = (UInt32) host_event_handler;
    bridge -> display_width = width;
   	bridge -> display_height = height;
    bridge -> display_pitch = pitch;
    bridge -> param1 = (unsigned long)name;
    bridge -> param2 = (unsigned long)Ftr;
    bridge -> flags = BF_UNZIP;
    
    if (res=ydk_PnoCall ( &pno, bridge )) {DEBUGS("Error while unzipping");DEBUGU32(res);return 4;}
    MemPtrFree(bridge);
	//unload pno
	res=ydk_PnoUnload ( &pno );    		
    if (res) {DEBUGS("Error while loading PNO(ZIP), probably not enough free ram.");}
   // rearrange memory
	garbage_collect();
	 		
	MemHandleUnlock(memh);
	DmReleaseResource(memh);
}



UInt32 LaunchPNO()
{
	LoadZippedModuleFromCard(/*PNO_ZDOOM*/ "zdoomz.zdk");

	return 0;								
}

Err CollapseDisplayResizedEventCallback( SysNotifyParamType *notifyParamsP )
{
	//uniquely post back a winDisplayChangedEvent in response to the notification
	//PinMgr 1.0 only posts the notification while PinMgr 1.1 posts back. By doing it this way 
	//your code need only handle the winDisplayChangedEvent
	EventType eventToAdd;
	
	MemSet(&eventToAdd, sizeof(EventType), 0);

	eventToAdd.eType = (eventsEnum) winDisplayChangedEvent;
	
	EvtAddUniqueEventToQueue(&eventToAdd, 0, true);
	
	return 0;
}

UInt32 AppMain(void)
{
	UInt32 res = errNone;
	UInt16 pitch, width, height;
	UInt16 devErr, compErr, error;
	UInt32 deviceID, companyID, version;
	UInt16 cardNo;
	LocalID dbID;
	unsigned short val;
	char c;
	
	if ( ! hw_set16bpp() )
	{
		DEBUGS ( "Couldn't set to 16 bit colour!" );
		return res;
	}

	g_launch_pno = 0;
	p_video_x = p_video_y = 320;
	p_video_rot = 0;
	device_type = 0;

	SysLibFind("Net.lib", &netLibRefNum);
   	
	{
		UInt32 pitch, width, height;
		hw_guess_screen_dimensions ( &width, &height, &pitch );
		if (/*(pitch>320)||(width>320)||(height>320)*/hw_flags & hf_pin_present)
		{
			device_type=_HIRES_PLUS_DEVICE_;
		}
	}
	    if (is_zodiac()) {device_type|=_ZODIAC_DEVICE_;}
   		if (is_clie()) device_type|=_CLIE_DEVICE_;
   		if (is_treo()) device_type|=_TREO_DEVICE_;
		
			/*Setup dir if necessary*/
			CheckDirs();
			/*Register file type for emu module*/
			VFSUnregisterDefaultDirectory ( ".zdk", expMediaType_Any);
		    VFSRegisterDefaultDirectory ( ".zdk", expMediaType_Any, VFSPATH );
		
			/* DEFAULT PREF */
			InitPrefDefaultValues();
									
			/* LOAD PREF IF AVAILABLE WITH GOOD VERSION NUMBER*/
			val=1;
			if (PrefGetAppPreferences(CREATORID,0x0001,(void*)&c,(unsigned short*)&val,(unsigned char)1)==VERSION_DB);
			{
				val=sizeof(g_zdoomz_pref);
				PrefGetAppPreferences(CREATORID,0x0001,(void*)&g_zdoomz_pref,(unsigned short*)&val,(unsigned char)1);
			}
			
			/* init random seed */
			SysRandom(TimGetTicks());
			
	error = FtrGet(pinCreator, pinFtrAPIVersion, &version);
	if (!error && version)
	{
		//PINS exists
		g_CollapseAPIVersion = version;
	}
	else
	{
		if ((error = SysLibFind(sonySysLibNameSilk, &SonySilkLibRefNum)))
		{
			if (error == sysErrLibNotFound) 
			{ /* couldn't find lib */ 
				error = SysLibLoad( 'libr', sonySysFileCSilkLib, &SonySilkLibRefNum );
				g_CollapseAPIVersion = 0;
			}
		}

		if (!error ) 
		{ 
			error = FtrGet(sonySysFtrCreator, sonySysFtrNumVskVersion, &version); 

			if (error)
			{ /* Version 1 is installed only resize is available */
				if (SilkLibOpen(SonySilkLibRefNum) == errNone)
				{
//					SilkLibEnableResize(SonySilkLibRefNum);
//					SilkLibDisableResize(SonySilkLibRefNum);
					g_CollapseAPIVersion = sonyVersion1;
				}
			}
			else
			{ /* Version 2 or up is installed Some added features are available */
				if (VskOpen(SonySilkLibRefNum) == errNone)
				{
//					VskSetState(SonySilkLibRefNum, vskStateEnable, (vskResizeVertically|vskResizeHorizontally));
//					VskSetState(SonySilkLibRefNum, vskStateEnable, 0);
					g_CollapseAPIVersion = sonyVersion2;
				}
			}
		}
	}

	if (g_CollapseAPIVersion > 0)
	{
		if (SysCurAppDatabase( &cardNo, &dbID ) == errNone)
			SysNotifyRegister( cardNo, dbID, sysNotifyDisplayResizedEvent, CollapseDisplayResizedEventCallback, sysNotifyNormalPriority, NULL );
	}
	else if (g_CollapseAPIVersion < 0)
	{
		if (SysCurAppDatabase( &cardNo, &dbID ) == errNone)
			SysNotifyRegister( cardNo, dbID, sysNotifyDisplayChangeEvent, CollapseDisplayResizedEventCallback, sysNotifyNormalPriority, NULL );
	}
	
	FrmGotoForm( BlankForm );
												
	AppEventLoop();
	AppStop();

	return res;
}

/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version is meet your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags that indicate if the application 
 *                                UI is initialized.
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
	UInt32 romVersion;

	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion)
	{
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
		{
			FrmAlert (RomIncompatibleAlert);
		
			// Palm OS 1.0 will continuously relaunch this app unless we switch to 
			// another safe one.
			if (romVersion <= kPalmOS10Version)
			{
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		
		return sysErrRomIncompatible;
		}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    NativeFuncType* entry;
    UInt32 res;
	
	if (cmd == sysAppLaunchCmdNormalLaunch) 
	{	
		res=0;
		if (!RomVersionCompatible(kOurMinVersion, launchFlags))
		{
			res=AppMain();
		}
	}

	return res;
}
